==> dna-0.5.1.tar.gz: lib/src/dynamic_library.dart <==
//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:typed_data';

import 'coder.dart';
import 'native.dart';
import 'meta.dart';
import 'model/library_meta.dart';
import 'model/method_meta.dart';

class DynamicLibrary<T> {
  final int STACK_SIZE = 256;

  Type type;
  LibraryMeta library;

  DynamicLibrary([Resolver resolver]) {
    this.type = T;

    assert(type != dynamic);
    //print('type $type');

    library = getLibrary(this.type, resolver);
    //FIXME assert(library.pointer != NULL_POINTER);
    //print('lib pointer ${library.pointer}');
  }

  dynamic noSuchMethod(Invocation invocation) {
    //print('method ${MirrorSystem.getName(invocation.memberName)}');

    var methodMeta = getMethod(invocation, library);
    assert(methodMeta.pointer != NULL_POINTER);
    //print('method pointer ${methodMeta.pointer}');

    //Array with TypedData object for params
    var paramsData = new List<TypedData>(invocation.positionalArguments.length);

    var vm = createCallVM(STACK_SIZE);
    assert(vm != NULL_POINTER);

    var nativeParams = before(invocation, methodMeta, paramsData);
    pushParameters(vm, methodMeta.types, nativeParams);
    var result = call(methodMeta, vm);
    after(invocation, methodMeta, paramsData);

    freeCallVM(vm);
    return result;
  }

  List<Object> before(
      Invocation invocation, MethodMeta methodMeta, List<TypedData> data) {
    var nativeParams = new List<Object>();

    var index = 0;
    for (var param in methodMeta.params) {
      var argument = invocation.positionalArguments[index];

      if (param.list) {
        if (param.out) {
          data[index] = encodeList(argument.value, param.type);
          argument = data[index];
        } else {
          argument = encodeList(argument, param.type);
        }
      }

      switch (param.type) {
        case C_STRING:
          {
            if (param.out) {
              data[index] = new Uint8List((argument.value as String).length);
              argument = data[index];
            } else {
              argument = new Uint8List.fromList((argument as String).codeUnits);
            }
          }
          break;
        case C_STRUCT:
          {
            if (argument == null) {
              argument = 0;
              break;
            }

            if (param.out) {
              //data[index] = encodeStruct(argument.value);
              argument.data = encodeStruct(argument.value);
              argument = argument.data;
            } else {
              argument = encodeStruct(argument);
            }
          }
          break;
        case C_POINTER:
          {
            if (param.ref) {
              argument = argument.value;
            }
          }
          break;
      }

      nativeParams.add(argument);
      index++;
    }

    return nativeParams;
  }

  void after(
      Invocation invocation, MethodMeta methodMeta, List<TypedData> data) {
    var index = 0;
    for (var param in methodMeta.params) {
      var argument = invocation.positionalArguments[index];

      if (param.list) {
        if (param.out) {
          argument.value = decodeList(data[index], param.type);
        }
      }

      switch (param.type) {
        case C_STRING:
          {
            if (param.out) {
              Uint8List list = (data[index] as Uint8List);
              argument.value = new String.fromCharCodes(list);
            }
          }
          break;
        case C_STRUCT:
          {
            if (argument == null) {
              argument = 0;
              break;
            }
            if (param.out) {
              //print('out ${argument.data} ${argument.data.toList().length}');
              argument.value = decodeStruct(argument.data, argument.value);
            }
          }
          break;
      }
      index++;
    }
  }

  dynamic call(MethodMeta methodMeta, int vm) {
    var result;
    if (methodMeta.returnParam.type == VOID) {
      callFunction(vm, methodMeta.pointer, methodMeta.returnParam.type);
    } else {
      result =
          callFunction(vm, methodMeta.pointer, methodMeta.returnParam.type);
    }

    return result;
  }
}

==> dna-0.5.3.tar.gz: lib/src/dynamic_library.dart <==
//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:typed_data';

import 'coder.dart';
import 'native.dart';
import 'meta.dart';
import 'model/library_meta.dart';
import 'model/method_meta.dart';

class DynamicLibrary<T> {
  final int STACK_SIZE = 256;

  Type _type;
  LibraryMeta _library;

  DynamicLibrary([Resolver resolver]) {
    _type = T;

    assert(_type != dynamic, 'dynamic is not supported');
    //print('_type $_type');

    _library = getLibrary(_type, resolver);    
    //print('lib pointer ${_library.pointer}');
  }

  dynamic noSuchMethod(Invocation invocation) {
    
    var methodMeta = getMethod(invocation, _library);
    assert(methodMeta.pointer != NULL_POINTER, 'cannot find method ${invocation.memberName}');
    //print('method pointer ${methodMeta.pointer}');

    //Array with TypedData object for params
    var paramsData = new List<TypedData>(invocation.positionalArguments.length);

    var vm = createCallVM(STACK_SIZE);
    assert(vm != NULL_POINTER, 'cannot create vm');

    var nativeParams = _before(invocation, methodMeta, paramsData);
    pushParameters(vm, methodMeta.types, nativeParams);
    var result = _call(methodMeta, vm);
    _after(invocation, methodMeta, paramsData);

    freeCallVM(vm);
    return result;
  }

  List<Object> _before(
      Invocation invocation, MethodMeta methodMeta, List<TypedData> data) {
    var nativeParams = new List<Object>();

    var index = 0;
    for (var param in methodMeta.params) {
      var argument = invocation.positionalArguments[index];

      if (param.list) {
        if (param.out) {
          data[index] = encodeList(argument.value, param.type);
          argument = data[index];
        } else {
          argument = encodeList(argument, param.type);
        }
      }

      switch (param.type) {
        case C_STRING:
          {
            if (param.out) {
              data[index] = new Uint8List((argument.value as String).length);
              argument = data[index];
            } else {
              argument = new Uint8List.fromList((argument as String).codeUnits);
            }
          }
          break;
        case C_STRUCT:
          {
            if (argument == null) {
              argument = 0;
              break;
            }

            if (param.out) {
              //data[index] = encodeStruct(argument.value);
              argument.data = encodeStruct(argument.value);
              argument = argument.data;
            } else {
              argument = encodeStruct(argument);
            }
          }
          break;
        case C_POINTER:
          {
            if (param.ref) {
              argument = argument.value;
            }
          }
          break;
      }

      nativeParams.add(argument);
      index++;
    }

    return nativeParams;
  }

  void _after(
      Invocation invocation, MethodMeta methodMeta, List<TypedData> data) {
    var index = 0;
    for (var param in methodMeta.params) {
      var argument = invocation.positionalArguments[index];

      if (param.list) {
        if (param.out) {
          argument.value = decodeList(data[index], param.type);
        }
      }

      switch (param.type) {
        case C_STRING:
          {
            if (param.out) {
              Uint8List list = (data[index] as Uint8List);
              argument.value = new String.fromCharCodes(list);
            }
          }
          break;
        case C_STRUCT:
          {
            if (argument == null) {
              argument = 0;
              break;
            }
            if (param.out) {
              //print('out ${argument.data} ${argument.data.toList().length}');
              argument.value = decodeStruct(argument.data, argument.value);
            }
          }
          break;
      }
      index++;
    }
  }

  dynamic _call(MethodMeta methodMeta, int vm) {
    var result;
    if (methodMeta.returnParam.type == VOID) {
      callFunction(vm, methodMeta.pointer, methodMeta.returnParam.type);
    } else {
      result =
          callFunction(vm, methodMeta.pointer, methodMeta.returnParam.type);
    }

    return result;
  }
}
